import re
from nssurge_utils.types import SurgeConfigSections


def get_sep_delim(section):
    match section:
        case 'Rule':
            value_delimiter = ','
            name_value_separator = ','
            # 1. normal rule
            # # >> Tencent AIA
            # IP-CIDR,162.14.0.0/18,no-resolve
            # 2. rule set
            # RULE-SET,https://ruleset.skk.moe/List/non_ip/global.conf,Proxy
        case 'URL Rewrite':
            # examples
            # cSpell:disable
            # ^http://\[::\] http://127.0.0.1 302
            # ^https?:\/\/cdn\.fivecdm\.com\/cr\/ - reject
            # cSpell:enable
            value_delimiter = ' '
            name_value_separator = ' '
        case _:
            value_delimiter = ','
            name_value_separator = '='
    return value_delimiter, name_value_separator


def parse_section(
    line: str, section: SurgeConfigSections | None = None
) -> tuple[str, list[str]] | None:
    comment_markers = '#'
    if line.startswith(comment_markers):
        return None
    value_delimiter, name_value_separator = get_sep_delim(section)
    # first partition line using name_value_separator
    name, value = line.partition(name_value_separator)[::2]
    name = name.strip()
    value = value.strip()
    # then partition value using value_delimiter
    values = [x.strip() for x in value.split(value_delimiter)]
    return name, values


def unparse_section(
    name: str, values: list[str], section: SurgeConfigSections | None = None
) -> str:
    value_delimiter, name_value_separator = get_sep_delim(section)
    value_str = f'{value_delimiter} '.join(values)
    section_str = f"{name}{name_value_separator if name_value_separator == ',' else f' {name_value_separator} '}{value_str}"
    return section_str



from os import PathLike
from nssurge_utils.types import SurgeConfigSections
from typing import Callable, Generator, List, Literal, Tuple


def file_line_generator(file_path: PathLike) -> Generator[str, None, None]:
    with open(file_path, 'r') as file:
        for line in file:
            yield line


def partition_lines(
    line_gen: Generator[str, None, None],
    is_first_marker: Callable[[str], bool],
    is_second_marker: Callable[[str], bool],
    include_markers: Tuple[bool, bool] = (False, False),
) -> tuple[list[str], list[str], list[str]]:
    lines_before = []
    lines_in_between = []
    lines_after = []
    # before_in_between = True
    # in_between = False
    location: Literal['before', 1, 'between', 2, 'after'] = 'before'

    for line in line_gen:
        match location:
            case 'before':
                if is_first_marker(line):
                    if include_markers[0]:
                        lines_in_between.append(line)
                    else:
                        lines_before.append(line)
                    location = 1
                else:
                    lines_before.append(line)
            case 1 | 'between':
                if is_second_marker(line):
                    if include_markers[1]:
                        lines_in_between.append(line)
                    else:
                        lines_after.append(line)
                    location = 2
                else:
                    lines_in_between.append(line)
                    location = 'between'
            case 2 | 'after':
                lines_after.append(line)
                location = 'after'
    return lines_before, lines_in_between, lines_after


def is_section_marker(line: str, section: SurgeConfigSections | None = None) -> bool:
    import re

    if section is None:
        return re.match(r'^\[(.*)\]', line) is not None
    return re.match(r'^\[' + section + r'\]', line) is not None


def merge_file_partitions(
    input_file: PathLike,
    target_file: PathLike,
    is_first_marker_input: Callable[[str], bool],
    is_second_marker_input: Callable[[str], bool],
    is_first_marker_output: Callable[[str], bool],
    is_second_marker_output: Callable[[str], bool],
    prepend: bool = True,
    replace: bool = False,
    output_file: PathLike | None = None,
) -> None:
    # Extract lines from input file
    _, extracted_input_lines, _ = partition_lines(
        file_line_generator(input_file),
        is_first_marker_input,
        is_second_marker_input,
        include_markers=(False, False),
    )

    # Process output file
    target_lines = partition_lines(
        file_line_generator(target_file),
        is_first_marker_output,
        is_second_marker_output,
        include_markers=(False, False),
    )
    if replace:
        replaced_lines = extracted_input_lines
    elif prepend:
        replaced_lines = extracted_input_lines + target_lines[1]
    else:
        replaced_lines = target_lines[1] + extracted_input_lines
    output_lines = target_lines[0] + replaced_lines + target_lines[2]
    if output_file is None:
        output_file = target_file
    with open(output_file, 'w') as file:
        file.writelines(output_lines)


def put_to_first_if_exist(l: list, members: set) -> list:
    """
    For Proxy Group rule values

    write a python function put_to_first_if_exist(l: list, members: set) -> list
    That check if any member of  members exists in l, if yes, remove all other members in l if exists, and make sure to put the member to the first in the output list, and make sure there’s no duplicate of this member in the output list.

    the output list should contain at most 1 member, and if it contain a member, the member should be at index 0, and there’s no duplicate.
    """
    members_exist_in_l = {member for member in members if member in l}
    if len(members_exist_in_l) >= 1:
        selected_member = members_exist_in_l.pop()
        # remove any other members in l if exists
        l = [x for x in l if x not in members]
        # put the member to the first in the output list
        l.insert(0, selected_member)
        return l
    else:
        return l



#!/usr/bin/env python3
"""
Author : Xinyuan Chen <45612704+tddschn@users.noreply.github.com>
Date   : 2023-05-24
Purpose: Extract the [Proxy] section from a Surge config file
"""

import argparse
from functools import partial
from pathlib import Path
from pprint import pprint
from nssurge_utils.utils import (
    partition_lines,
    is_section_marker,
    file_line_generator,
    merge_file_partitions,
    put_to_first_if_exist,
)
from nssurge_utils.config import surge_config_sections, special_proxy_group_value
from nssurge_utils.parsers import parse_section, unparse_section


def get_args():
    """Get command-line arguments"""

    parser = argparse.ArgumentParser(
        description='Extract the [Proxy] section from a Surge config file',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    parser.add_argument('file', metavar='FILE', type=Path, help='Surge config file')
    parser.add_argument(
        '-s',
        '--section',
        metavar='SECTION',
        type=str,
        default='Proxy',
        choices=surge_config_sections,
        help='The section to extract',
    )
    parser.add_argument(
        '--add-to-proxy',
        metavar='FILE',
        type=Path,
        help='Add the extracted proxy config to the proxy section of file',
    )
    parser.add_argument(
        '--add-to-proxy-group',
        action='store_true',
        help='Add the extracted proxy names to the proxy group section of file',
    )
    parser.add_argument(
        '-o',
        '--output',
        metavar='FILE',
        type=Path,
        help='Output file, if not specified, modification will be done in-place',
    )

    args = parser.parse_args()
    if args.add_to_proxy_group and not args.add_to_proxy:
        parser.error(
            '--add-to-proxy-group requires --add-to-proxy to be specified as well'
        )
    if not args.output:
        args.output = args.add_to_proxy
    return args


def main():
    """Make a jazz noise here"""

    args = get_args()
    if args.add_to_proxy:
        merge_file_partitions(
            input_file=args.file,
            target_file=args.add_to_proxy,
            is_first_marker_input=partial(is_section_marker, section=args.section),
            is_second_marker_input=is_section_marker,
            is_first_marker_output=is_section_marker,
            is_second_marker_output=is_section_marker,
            output_file=args.output,
        )
        if args.add_to_proxy_group:
            _, extracted_lines_input, _ = partition_lines(
                file_line_generator(args.file),
                is_first_marker=partial(is_section_marker, section=args.section),
                is_second_marker=is_section_marker,
            )
            lines_before, extracted_lines_target, lines_after = partition_lines(
                file_line_generator(args.output),
                is_first_marker=partial(is_section_marker, section='Proxy Group'),
                is_second_marker=is_section_marker,
            )
            proxy_names = []
            for proxy_rule in extracted_lines_input:
                if (parsed := parse_section(proxy_rule)) is None:
                    continue
                proxy_name, _ = parsed
                proxy_names.append(proxy_name)

            new_proxy_group_rules = []
            for proxy_group_rule in extracted_lines_target:
                if (parsed := parse_section(proxy_group_rule)) is None:
                    continue
                proxy_group_name, proxy_group_values = parsed
                new_proxy_group_rule = unparse_section(
                    proxy_group_name,
                    put_to_first_if_exist(
                        proxy_names + proxy_group_values, special_proxy_group_value
                    ),
                    section='Proxy Group',
                )
                new_proxy_group_rules.append(new_proxy_group_rule + '\n')
            # write lines_before, new_proxy_group_rules, lines_after to output_file
            with open(args.output, 'w') as f:
                f.writelines(lines_before)
                f.writelines(new_proxy_group_rules)
                f.writelines(lines_after)
        return

    _, extracted_lines_input, _ = partition_lines(
        file_line_generator(args.file),
        is_first_marker=partial(is_section_marker, section=args.section),
        is_second_marker=is_section_marker,
    )
    print(''.join(extracted_lines_input), end='')


if __name__ == '__main__':
    main()



from typing import Literal, get_args
from nssurge_utils.types import SurgeConfigSections

# Retrieve the valid options from SurgeConfigSections
# valid_sections = get_args(SurgeConfigSections)
surge_config_sections: tuple[str] = get_args(SurgeConfigSections)

# Iterate through all valid values
# for section in valid_sections:
#     print(section)

# must be the first ones
special_proxy_group_value = {'select', 'url-test'}



# [General]
# [Replica]
# [Proxy]
# [Proxy Group]
# [Rule]
# [Host]
# [URL Rewrite]
# [Script]
# [MITM]

from typing import Literal


SurgeConfigSections = Literal[
    'General',
    'Replica',
    'Proxy',
    'Proxy Group',
    'Rule',
    'Host',
    'URL Rewrite',
    'Script',
    'MITM',
]
